# SPDX-FileCopyrightText: 2023-present rmrighes <rmrighes@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT
"""
File __about__.py responsible for yajaw package meta information.
"""


__version__ = "0.1.0.dev1"
